/*
  Extension       LinuxMouse
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Linux TSE, TSE version undetermined.
                  Needs advanced Linux terminal.
  Version         v0.0.0.2   1 May 2024.

  !!! In Development. See DISCLAIMERS. !!!

  This extension displays the mouse's position and buttons status
  when clicking in the editing text.

  Nothing is displayed (yet) when a menu, prompt or list is open.
  Sometimes there are unfortunate side effects.

  Which buttons are pressed is shown as an 8 bit ButtonStatus.

  Known errors:
  - Often it does not distinguish between between the left and right button.
    I suspect/hope that this is a display error of the macro,
    and not a limitation of the the terminal's functionality.
  - After using the <Wheel Up/Down> plus <Escape> TSE's Help is opened.

  TODO
    SHOULD
    - Upon a mouse click move the text cursor to the mouse position.
    - Distinguish between left and right button.
    - Work in TSE menus, prompts, lists.
    - Mouse marking.
    COULD
    WOULD
    WONT

  DISCLAIMERS !!!

    This extension only works in Linux terminals that support certain
    newer/more advanced terminal protocols.
    In older/simpler Linux terminals this extension will not work,
    with as yet unknown consequences.

    The extension does not support column and row positions above 223.

    This is a development version!
    The extension is likely to fail with unknown consequences.
*/


/*
  T E C H N I C A L   B A C K G R O U N D   I N F O

  Context: X11 mouse tracking enabled in an up-to-date WSL 2.

  Using the EventLogger extension to monitor _AFTER_GETKEY_ was super helpful:

    I selected only the _AFTER_GETKEY_ event.

    It default already showed which keys TSE saw.

    Because a keystroke's time is logged with great precision,
    pausing between clicking revealed that TSE behaves as follows.

    A mouse button press or release generates 1 physical keystroke
    for key <F1> and adds 6 response characters to STDIN.

    TSE only processes the 1 physical keystroke for <F1>
    and does not process STDIN as keys until the next physical keystroke.

    At the next physical keystroke TSE processes STDIN followed by whatever the
    physical keystroke was.

    When processing STDIN TSE "ignores" or "receives no keys"
    the first 3 characters of the 6-character response string,
    which are always Chr(27) + "[M" (aka CSI M).

  Reading STDIN has no impact on processing keys.

  A button click generates two events: button Down and button Up.


  INSTALLATION

  Compile the macro and temporarily load it to test it.

*/


// Constants and semi-constants
string  CSI_M [3] = Chr(0x1B) + '[M'


// Global variables
string  mouse_properties      [3] = ''
string  msg        [MAXSTRINGLEN] = ''
integer num_ignorable_keys        = 0
integer num_mouse_status_displays = 0


proc ignore_current_key()
  BreakHookChain()
  Set(Key, -1)
end ignore_current_key


proc after_getkey()
  string stdin_queue [MAXSTRINGLEN] = ''

  if num_ignorable_keys
    ignore_current_key()
    num_ignorable_keys = num_ignorable_keys - 1
  elseif Query(Key) == <F1>
    fRead(_STDIN_, stdin_queue, MAXSTRINGLEN)

    if stdin_queue[1:3] == CSI_M
      mouse_properties          = stdin_queue[4:3]
      num_ignorable_keys        = 3
      num_mouse_status_displays = 4

      ignore_current_key()

      msg = Format(' ButtonStatus=',
                   Asc(mouse_properties[1]):8:'0':2;
                   'MouseColumn=',
                   Asc(mouse_properties[2]) - 32;
                   'MouseRow=',
                   Asc(mouse_properties[3]) - 32,
                   '.')
      msg = Format(msg:-52)
    endif
  endif

  if num_mouse_status_displays
    num_mouse_status_displays = num_mouse_status_displays - 1
    PutStrAttrXY(Query(ScreenCols) - Length(msg),
                 Query(ScreenRows) - 3,
                 msg,
                 '',
                 Color(bright yellow ON blue))
    msg = ''
  endif
end after_getkey


proc WhenPurged()
  fWrite(_STDOUT_, Chr(27) + "[?1000l")   // Set X11 mouse tracking mode OFF.
end WhenPurged


proc WhenLoaded()
  fWrite(_STDOUT_, Chr(27) + "[?1000h")   // Set X11 mouse tracking mode ON.

  Hook(_AFTER_GETKEY_     , after_getkey)
  Hook(_ON_ABANDON_EDITOR_, WhenPurged  ) // Stop mouse tracking after editor.
end WhenLoaded


proc Main()
end Main

